/*
 Local Rank Distance Sequence Aligner 1.0
 
 This algorithm is based on a positional inverted index that improves the efficiency.
 Aligns a set of reads against a reference DNA sequence using Local Rank Distance.
 
 Copyright (C) 2013  Liviu P. Dinu, Radu Tudor Ionescu, Alexandru I. Tomescu
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or any
 later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <fstream>
#include <cmath>
#include <stdlib.h>
#include <time.h>
#include <limits>
#include <string>
#include <string.h> // for linux
#include <vector>
#include <unordered_map>

using namespace std;

class LRDAligner
{
private:
    char **reads;
    char **readNames;
    
    char *referenceSeq;
    char referenceSeqName[256];
    string reference;
    vector<vector<int> > referencePositions;

    int ngramSize, maxOffset;
    int readSize, readsCount;
    int referenceSize;
    
    int maxAllowedDistance;
    int notFoundNgramsLimit;
    
    int maxNgramsCountPerRead;
    
    int **alignments;
    int *foundNgrams;
    int *countNgrams;
    
    char *tempRead;
    int *tempAlignment;
    
    void deleteData();

    int* alignReadAtIndex(int, int, int);
    
public:
    LRDAligner(char *, char *, char *);
    ~LRDAligner();
    
    int align(char *);
};

// Loads genomes based on a config file given as parameter
LRDAligner::LRDAligner(char *configFile, char *readsFile, char *referenceFile)
{
	int i, j;
    char ch;
    string dropLine;

    // Read data from config file
	ifstream cfgin;
	cfgin.open(configFile);
	if (!cfgin) 
	{
		cout<<"Unable to open file at path: "<<configFile<<endl;
		throw 0;
	}
    if (!(cfgin>>ngramSize))
	{
		cout<<"File at path : "<<configFile<<" is corrupted!"<<endl;
		cfgin.close();
		throw 0;
	}
    if (!(cfgin>>maxOffset))
	{
		cout<<"File at path : "<<configFile<<" is corrupted!"<<endl;
		cfgin.close();
		throw 0;
	}
    if (!(cfgin>>readsCount))
	{
		cout<<"File at path : "<<configFile<<" is corrupted!"<<endl;
		cfgin.close();
		throw 0;
	}
    if (!(cfgin>>readSize))
	{
		cout<<"File at path : "<<configFile<<" is corrupted!"<<endl;
		cfgin.close();
		throw 0;
	}
    if (!(cfgin>>maxAllowedDistance))
	{
		cout<<"File at path : "<<configFile<<" is corrupted!"<<endl;
		cfgin.close();
		throw 0;
	}
    if (!(cfgin>>notFoundNgramsLimit))
	{
		cout<<"File at path : "<<configFile<<" is corrupted!"<<endl;
		cfgin.close();
		throw 0;
	}
    cfgin.close();
    
    ifstream qin;
    qin.open(readsFile);
    if (!qin)
    {
        cout<<"Unable to open file at path: "<<readsFile<<endl;
        throw 0;
    }
    
    ifstream refin;
    refin.open(referenceFile);
    if (!refin)
    {
        cout<<"Unable to open file at path: "<<referenceFile<<endl;
        throw 0;
    }
    
    maxNgramsCountPerRead = readSize - ngramSize;
    
    // Get size of reference genome
    do
    {
        getline(refin,dropLine); // drop first line with sequence identifier
    } while (dropLine[0] != '>');
    for (i = 1; dropLine[i] != ' ' && dropLine[i] != '\n' && i < 256; i++)
        referenceSeqName[i-1] = dropLine[i];
    referenceSeqName[i-1] = '\0';
    
    for (referenceSize = 0; refin >> ch; referenceSize++);
    refin.close();
    refin.open(referenceFile);
    if (!refin)
    {
        cout<<"Unable to open file at path: "<<referenceFile<<endl;
        throw 0;
    }
    
	try
	{
		reads = NULL;
		reads = new char*[readsCount];
        for (i = 0; i < readsCount; i++)
        {
			reads[i] = NULL;
            reads[i] = new char[readSize + 1];
        }
        
        readNames = NULL;
        readNames = new char*[readsCount];
        for (i = 0; i < readsCount; i++)
        {
			readNames[i] = NULL;
            readNames[i] = new char[256];
        }
        
        tempRead = NULL;
        tempRead = new char[readSize];
        
        tempAlignment = NULL;
        tempAlignment = new int[2 + referenceSize];
        
		alignments = new int*[readsCount];
		for (i = 0; i < readsCount; i++)
		{
			alignments[i] = NULL;
		}
        
        referenceSeq = NULL;
        referenceSeq = new char[referenceSize + 1];
        
        foundNgrams = new int[referenceSize];
        countNgrams = new int[referenceSize];
        
        // Read data from reads (fastq) file
		for (i = 0; i < readsCount; i++)
		{
            do
            {
                getline(qin,dropLine); // drop first line with sequence identifier
            } while (dropLine[0] != '@');
            for (j = 1; dropLine[j] != ' ' && dropLine[j] != '\n' && j < 256; j++)
                readNames[i][j-1] = dropLine[j];
            readNames[i][j-1] = '\0';
            
            for (j = 0; j < readSize;)
                if (qin >> ch)
                {
                    if (ch == 'A' || ch == 'C' || ch =='G' || ch == 'T')
                    {
                        reads[i][j] = ch;
                    }
                    j++;
                }
                else
                {
					cout<<"File at path : "<<readsFile<<" is corrupted!"<<endl;
					throw 0;
				}
            reads[i][readSize] = '\0';
        }
        
        // Read data from reference sequence (fasta) file
        do
        {
            getline(refin,dropLine); // drop first line with sequence identifier
        } while (dropLine[0] != '>');
        
        for (i = 0; i < referenceSize;)
            if (refin >> ch)
            {
                if (ch == 'A' || ch == 'C' || ch =='G' || ch == 'T')
                {
                    referenceSeq[i] = ch;
                }
                i++;
            }
            else
            {
                cout<<"File at path : "<<referenceFile<<" is corrupted!"<<endl;
                throw 0;
            }
        referenceSeq[referenceSize] = '\0';
        
        reference = string(referenceSeq);
        referencePositions = vector<vector<int > >(referenceSize, vector<int>(0));
        
        qin.close();
        refin.close();
	}
	catch (...)
	{
		deleteData();
        qin.close();
        refin.close();
		throw;
	}
}

LRDAligner::~LRDAligner()
{
	int i;
	for (i = 0; i < readsCount; i++)
		delete[] reads[i];
    delete[] reads;
    reads = NULL;
    
    for (i = 0; i < readsCount; i++)
		delete[] readNames[i];
    delete[] readNames;
    readNames = NULL;
    
    delete[] tempRead;
    tempRead = NULL;
    
    delete[] tempAlignment;
    tempAlignment = NULL;
    
	for (i = 0; i < readsCount; i++)
		delete[] alignments[i];
	delete[] alignments;
	alignments = NULL;
    
    delete[] foundNgrams;
    foundNgrams = NULL;
    
    delete[] countNgrams;
    countNgrams = NULL;
    
	delete[] referenceSeq;
	referenceSeq = NULL;
	readsCount = readSize = referenceSize = ngramSize = maxOffset = 0;
    maxNgramsCountPerRead = maxAllowedDistance = notFoundNgramsLimit = 0;
}

void LRDAligner::deleteData()
{
    int i;
	if (reads != NULL)
	{
		for (i = 0; i < readsCount; i++)
            delete[] reads[i];
        delete[] reads;
        reads = NULL;
	}
    if (readNames != NULL)
    {
        for (i = 0; i < readsCount; i++)
            delete[] readNames[i];
        delete[] readNames;
        readNames = NULL;
    }
    if (tempRead != NULL)
    {
        delete[] tempRead;
        tempRead = NULL;
    }
    if (tempAlignment != NULL)
    {
        delete[] tempAlignment;
        tempAlignment = NULL;
    }
	if (alignments != NULL)
	{
		for (i = 0; i < readsCount; i++)
            delete[] alignments[i];
        delete[] alignments;
        alignments = NULL;
	}
	if (referenceSeq != NULL)
	{
		delete[] referenceSeq;
		referenceSeq = NULL;
	}
    if (foundNgrams != NULL)
    {
        delete[] foundNgrams;
        foundNgrams = NULL;
    }
}

// Aligns a read against the reference DNA sequence
int* LRDAligner::alignReadAtIndex(int readIndex, int useReverseComplement, int bestDist)
{
    int i, j, k, count, maxCount;
    
    if (useReverseComplement)
    {
        char complement1, complement2;
        for (i = 0; i < readSize / 2; i++)
        {
            complement1 = reads[readIndex][i];
            complement2 = reads[readIndex][readSize - 1 - i];
            
            if (complement1 == 'A') complement1 = 'T';
            else if (complement1 == 'C') complement1 = 'G';
            else if (complement1 == 'G') complement1 = 'C';
            else if (complement1 == 'T') complement1 = 'A';
            
            if (complement2 == 'A') complement2 = 'T';
            else if (complement2 == 'C') complement2 = 'G';
            else if (complement2 == 'G') complement2 = 'C';
            else if (complement2 == 'T') complement2 = 'A';
            
            reads[readIndex][i] = complement2;
            reads[readIndex][readSize - 1 - i] = complement1;
        }
        
        if (readSize % 2 == 1)
        {
            complement1 = reads[readIndex][i];
            
            if (complement1 == 'A') complement1 = 'T';
            else if (complement1 == 'C') complement1 = 'G';
            else if (complement1 == 'G') complement1 = 'C';
            else if (complement1 == 'T') complement1 = 'A';
            
            reads[readIndex][i] = complement1;
        }
    }
    
    // Build positional inverted index for the read
    unordered_map<string, vector<int> > ngramsHash;
    string s(reads[readIndex]);
    
    for (i = 0; i <= readSize - ngramSize; i++)
    {
        string ngram = s.substr(i, ngramSize);
        if (ngramsHash.count(ngram) > 0)
        {
            ngramsHash[ngram].push_back(i);
        }
        else
        {
            vector<int> positions;
            positions.push_back(i);
            ngramsHash[ngram] = positions;
        }
    }
    
    for (k = 0; k <= referenceSize - ngramSize; k++)
    {
        foundNgrams[k] = (ngramsHash.count(reference.substr(k, ngramSize)) > 0) ? 1 : 0;
        referencePositions[k] = (foundNgrams[k]) ? ngramsHash[reference.substr(k, ngramSize)] : vector<int>(0);
    }
    
    count = 0;
    for (k = 0; k <= readSize - ngramSize; k++)
        if (foundNgrams[k]) count++;
    
    countNgrams[0] = count;
    maxCount = count;
    
    for (k = 1; k <= referenceSize - readSize; k++)
    {
        if (foundNgrams[k - 1]) count--;
        if (foundNgrams[k + readSize - ngramSize]) count++;

        countNgrams[k] = count;
        if (countNgrams[k] > maxCount) maxCount = countNgrams[k];
    }
    
    /*
    cout << "Ngrams in hash of read " << readIndex << " : {";
    for(tr1::unordered_map<string, vector<int> >::iterator it = ngramsHash.begin(); it != ngramsHash.end(); it++)
        cout << it->first << ",";
    cout << "}" << endl;
    cout << "There are " << ngramsHash.size() << " in total." << endl;
     */
    
    tempAlignment[0] = 1;
    tempAlignment[1] = bestDist;
    tempAlignment[2] = -1;
    
    int minPos, midPos, maxPos;
    int dist, minOffset, offset;
    
    maxCount -= notFoundNgramsLimit;

    for (k = 0; k <= referenceSize - readSize; k++)
        if (countNgrams[k] >= maxCount && (maxNgramsCountPerRead - countNgrams[k]) * maxOffset < bestDist)
        {
            dist = 0;
            for (i = 0; i <= readSize - ngramSize; i++)
            {
                if (dist > bestDist) i = readSize;
                else
                {
                    if (foundNgrams[k + i])
                    {
                        minOffset = maxOffset;
                        vector<int> positions = referencePositions[k + i];
                        
                        /*
                        for(j = 0; j < positions.size(); j++)
                        {
                            offset = (int)abs(positions[j] - i);
                            if (offset < minOffset) minOffset = offset;
                        }*/
                        
                        minPos = 0;
                        maxPos = positions.size() - 1;
                        
                        while (maxPos - minPos > 1)
                        {
                            midPos = (minPos + maxPos) >> 1;
                            if (i == positions[midPos])
                            {
                                minPos = midPos;
                                maxPos = midPos;
                            }
                            else if (i < positions[midPos]) maxPos = midPos;
                            else minPos = midPos;
                        }
                        for(j = minPos; j <= maxPos; j++)
                        {
                            offset = (int)abs(positions[j] - i);
                            if (offset < minOffset) minOffset = offset;
                        }
                        
                        dist += minOffset;
                    }
                    else dist += maxOffset;
                }
            }
            
            if (dist < bestDist)
            {
                bestDist = dist;
                tempAlignment[0] = 1;
                tempAlignment[1] = bestDist;
                tempAlignment[2] = k;
            }
            else if (dist == bestDist)
            {
                if (tempAlignment[2] >= 0) tempAlignment[0]++;
                tempAlignment[1 + tempAlignment[0]] = k;
            }
        }

    int *alignment = NULL;
    try
    {
        alignment = new int[2 + tempAlignment[0]];
        for (i = 0; i < 2 + tempAlignment[0]; i++)
            alignment[i] = tempAlignment[i];
    }
    catch (...)
    {
        return NULL;
    }
    return alignment;
}

// Aligns all reads against a reference DNA sequence using LRD and writes the results in the file given as parameter
int LRDAligner::align(char *file)
{
	if (readsCount < 0) return 0;
	int i, j;

    int maxPossibleDistance = maxOffset * (readSize - ngramSize + 1);
    int maxDistance = (maxAllowedDistance < maxPossibleDistance) ? maxAllowedDistance : maxPossibleDistance;

    int maxAcceptableOffset = (maxOffset < ngramSize) ? maxOffset : ngramSize;
    int maxAcceptableDistance = (maxAllowedDistance < maxAcceptableOffset * (readSize - ngramSize + 1)) ? maxAllowedDistance : maxAcceptableOffset * (readSize - ngramSize + 1);
    
	for (i = 0; i < readsCount; i++)
	{
        int *alignment;
        int *reverseComplementAlignment;
        
        alignment = alignReadAtIndex(i, 0, maxDistance);
        
        if (alignment[1] > maxAcceptableDistance)
        {
            strcpy(tempRead, reads[i]);
            tempRead[readSize] = '\0';
            
            reverseComplementAlignment = alignReadAtIndex(i, 1, alignment[1]);
            
            if (alignment[1] <= reverseComplementAlignment[1])
            {
                alignments[i] = alignment;
                strcpy(reads[i], tempRead);
                reads[i][readSize] = '\0';
            }
            else
            {
                alignments[i] = reverseComplementAlignment;
            }
        }
        else
        {
            alignments[i] = alignment;
        }
        
        // if (i % 50 == 0) cout << "Aligned read at index " << i << endl;
	}
    
    // The result is written in the fout file
    ofstream fout;
	fout.open(file);
	if (!fout)
	{
		cout << "Unable to open file at path: " << file << endl;
		return 0;
	}
    
    fout << "@HD\tVN:1.4\tSO:unsorted" << endl;
    fout << "@SQ\tSN:" << referenceSeqName << "\tLN:" << referenceSize << endl;
	for (i = 0; i < readsCount; i++)
	{
		for (j = 2; j < 2 + alignments[i][0]; j++)
            if (alignments[i][2] >= 0)
            {
                // print QNAME FLAG RNAME
                fout << readNames[i] << "\t" << 0 << "\t" << referenceSeqName << "\t";
                // print POS
                fout << (alignments[i][j] + 1) << "\t";
                // print MAPQ
                fout << int(254 * ((double)alignments[i][1] / maxPossibleDistance)) << "\t";
                // print CIGAR RNEXT PNEXT TLEN
                fout << "0\t*\t0\t0\t";
                // print SEQ QUAL
                fout << reads[i] << "\t*\t";
                // print AS:i:<LRD>
                fout << "AS:i:" << alignments[i][1] << endl;
            }
            else
            {
                // print QNAME FLAG RNAME
                fout << readNames[i] << "\t" << 0 << "\t*\t0\t0\t";
                // print CIGAR RNEXT PNEXT TLEN
                fout << "*\t*\t0\t0\t";
                // print SEQ QUAL
                fout << reads[i] << "\t*" << endl;
            }
	}
	fout.close();
	return 1;
}

int main(int argc, char *argv[])
{
    if (argc != 5)
    {
        cout<<"Local Rank Distance Sequence Aligner 1.0"<<endl;
        cout<<"USAGE: ./lrda <config_file> <reads_file.fastq> <reference_sequence_file.fasta> <output_file.sam>"<<endl;
        cout<<"The format of the config file :"<<endl;
        cout<<"LINE   1\t<n_gram_size> <max_offset>"<<endl;
        cout<<"LINE   2\t<reads_count> <reads_size>"<<endl;
        cout<<"LINE   3\t<distance_threshold>"<<endl;
        cout<<"LINE   4\t<not_found_ngrams_limit_over_min>"<<endl;
        return 1;
    }
    else 
    {
        LRDAligner *lrd;
        try 
        {
            lrd = new LRDAligner(argv[1], argv[2], argv[3]);
            if (!lrd->align(argv[4])) return 1;
            
            delete lrd;
        }
        catch(...) 
        {
            return 1;
        }  	
    }
    return 0;
}
