/*
 Local Rank Distance Sequence Aligner 1.0
 
 This algorithm is based on a positional inverted index that improves the efficiency.
 Aligns a set of reads against a reference DNA sequence using Local Rank Distance.
 
 Copyright (C) 2013  Liviu P. Dinu, Radu Tudor Ionescu, Alexandru I. Tomescu
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or any
 later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;
import java.util.*;

class LRDAligner
{
    private String[] reads;
    private String[] readNames;

    private String reference;
    private String referenceSeqName;

    private ArrayList<ArrayList<Integer> > referencePositions;

    private int ngramSize, maxOffset;
    private int averageReadSize, readsCount;

    private int maxAllowedDistance;
    private int notFoundNgramsLimit;

    private int[][] alignments;
    private boolean[] foundNgrams;
    private int[] countNgrams;

    private String tempRead;
    private int[] tempAlignment;

    // Loads genomes based on a config file given as parameter
    public LRDAligner(String configFile, String readsFile, String referenceFile)
    {
        int i;
        String line;

        try
        {
            // Read the data from config file
            Scanner cfgin = new Scanner(new File(configFile));
            
            ngramSize = cfgin.nextInt();
            maxOffset = cfgin.nextInt();
            readsCount = cfgin.nextInt();
            averageReadSize = cfgin.nextInt();
            maxAllowedDistance = cfgin.nextInt();
            notFoundNgramsLimit = cfgin.nextInt();
            
            cfgin.close();

            // Read data from reference sequence (fasta) file
            BufferedReader refin = new BufferedReader(new InputStreamReader(new FileInputStream(referenceFile), "UTF-8"));
            
            do
            {
                line = refin.readLine(); // drop first line with sequence identifier
            } while (line != null && line.charAt(0) != '>');
            
            referenceSeqName = (line.indexOf(' ') >= 0) ? line.substring(1, line.indexOf(' ')) : line.substring(1);
            
            reference = new String();
            
            for (line = refin.readLine(); line != null; line = refin.readLine())
            {
                line = line.toUpperCase();
                reference = reference.concat(line);
            }
            refin.close();
            
            foundNgrams = new boolean[reference.length()];
            countNgrams = new int[reference.length()];
            
            // Read the DNA reads
            BufferedReader qin = new BufferedReader(new InputStreamReader(new FileInputStream(readsFile), "UTF-8"));
            
            reads = new String[readsCount];
            readNames = new String[readsCount];
            
            tempAlignment = new int[2 + reference.length()];
            alignments = new int[readsCount][];
            
            // Read data from reads (fastq) file
            for (i = 0; i < readsCount; i++)
            {
                do
                {
                    line = qin.readLine(); // drop first line with sequence identifier
                } while (line != null && line.charAt(0) != '@');
                
                readNames[i] = (line.indexOf(' ') >= 0) ? line.substring(1, line.indexOf(' ')) : line.substring(1);
                
                reads[i] = new String();

                for (line = qin.readLine(); line != null && line.charAt(0) != '+'; line = qin.readLine())
                {
                    line = line.toUpperCase();
                    reads[i]  = reads[i].concat(line);
                }
            }
            qin.close();
        }
        catch (Exception e)
        {
            System.out.println(e);
            System.exit(1);
        }
    }
    
    String reverseComplementedString(String sequence)
    {
        String reverseComplement = sequence.replace('A','t');
        reverseComplement = reverseComplement.replace('C','g');
        reverseComplement = reverseComplement.replace('G','c');
        reverseComplement = reverseComplement.replace('T','a');
        
        reverseComplement = new StringBuilder(reverseComplement).reverse().toString().toUpperCase();
        return reverseComplement;
    }

    // Aligns a read against the reference DNA sequence
    int[] alignReadAtIndex(int readIndex, boolean useReverseComplement, int bestDist)
    {
        int i, j, k, count, maxCount;
        
        String read = reads[readIndex];
        
        if (useReverseComplement)
        {
            read = reverseComplementedString(read);
        }
        
        // Build positional inverted index for the read
        HashMap<String, ArrayList<Integer> > ngramsHash = new HashMap<String, ArrayList<Integer> >();
        
        for (i = 0; i <= read.length() - ngramSize; i++)
        {
            String ngram = read.substring(i, i + ngramSize);
            
            ArrayList<Integer> positions = ngramsHash.get(ngram);
            
            if (positions != null)
            {
                positions.add(i);
            }
            else
            {
                positions = new ArrayList<Integer>();
                positions.add(i);
            }
            
            ngramsHash.put(ngram, positions);
        }
        
        referencePositions = new ArrayList<ArrayList<Integer> >();
        for (k = 0; k <= reference.length() - ngramSize; k++)
        {
            foundNgrams[k] = ngramsHash.containsKey(reference.substring(k, k + ngramSize));
            ArrayList<Integer> positions = (foundNgrams[k]) ? ngramsHash.get(reference.substring(k, k + ngramSize)) : new ArrayList<Integer>();
            referencePositions.add(positions);
        }
        
        count = 0;
        for (k = 0; k <= read.length() - ngramSize; k++)
            if (foundNgrams[k]) count++;
        
        countNgrams[0] = count;
        maxCount = count;
        
        for (k = 1; k <= reference.length() - read.length(); k++)
        {
            if (foundNgrams[k - 1]) count--;
            if (foundNgrams[k + read.length() - ngramSize]) count++;
            
            countNgrams[k] = count;
            if (countNgrams[k] > maxCount) maxCount = countNgrams[k];
        }
        
        tempAlignment[0] = 1;
        tempAlignment[1] = bestDist;
        tempAlignment[2] = -1;
        
        int minPos, midPos, maxPos;
        int dist, minOffset, offset;
        
        maxCount -= notFoundNgramsLimit;

        for (k = 0; k <= reference.length() - read.length(); k++)
            if (countNgrams[k] >= maxCount && (read.length() - ngramSize - countNgrams[k]) * maxOffset < bestDist)
            {
                dist = 0;
                for (i = 0; i <= read.length() - ngramSize; i++)
                {
                    if (dist > bestDist) i = read.length();
                    else
                    {
                        if (foundNgrams[k + i])
                        {
                            minOffset = maxOffset;
                            ArrayList<Integer> positions = referencePositions.get(k + i);
   
                            /*
                            for(j = 0; j < positions.size(); j++)
                            {
                                offset = (int)abs(positions[j] - i);
                                if (offset < minOffset) minOffset = offset;
                            }*/
                            
                            minPos = 0;
                            maxPos = positions.size() - 1;
                            
                            while (maxPos - minPos > 1)
                            {
                                midPos = (minPos + maxPos) >> 1;
                                if (i == positions.get(midPos))
                                {
                                    minPos = midPos;
                                    maxPos = midPos;
                                }
                                else if (i < positions.get(midPos)) maxPos = midPos;
                                else minPos = midPos;
                            }
                            for(j = minPos; j <= maxPos; j++)
                            {
                                offset = (int)Math.abs(positions.get(j) - i);
                                if (offset < minOffset) minOffset = offset;
                            }
                            
                            dist += minOffset;
                        }
                        else dist += maxOffset;
                    }
                }
                
                if (dist < bestDist)
                {
                    bestDist = dist;
                    tempAlignment[0] = 1;
                    tempAlignment[1] = bestDist;
                    tempAlignment[2] = k;
                }
                else if (dist == bestDist)
                {
                    if (tempAlignment[2] >= 0) tempAlignment[0]++;
                    tempAlignment[1 + tempAlignment[0]] = k;
                }
            }

        int[] alignment = new int[2 + tempAlignment[0]];
        for (i = 0; i < 2 + tempAlignment[0]; i++)
            alignment[i] = tempAlignment[i];
        
        return alignment;
    }

    // Aligns all reads against a reference DNA sequence using LRD and writes the results in the file given as parameter
    void align(String file)
    {
        if (readsCount < 0) return;
        int i, j;

        int maxPossibleDistance = maxOffset * (averageReadSize - ngramSize + 1);
        int maxDistance = (maxAllowedDistance < maxPossibleDistance) ? maxAllowedDistance : maxPossibleDistance;

        int maxAcceptableOffset = (maxOffset < ngramSize) ? maxOffset : ngramSize;
        int maxAcceptableDistance = (maxAllowedDistance < maxAcceptableOffset * (averageReadSize - ngramSize + 1)) ? maxAllowedDistance : maxAcceptableOffset * (averageReadSize - ngramSize + 1);
        
        for (i = 0; i < readsCount; i++)
        {
            int[] alignment;
            int[] reverseComplementAlignment;
            
            alignment = alignReadAtIndex(i, false, maxDistance);
            
            if (alignment[1] > maxAcceptableDistance)
            {
                reverseComplementAlignment = alignReadAtIndex(i, true, alignment[1]);
                
                if (alignment[1] <= reverseComplementAlignment[1])
                {
                    alignments[i] = alignment;
                }
                else
                {
                    alignments[i] = reverseComplementAlignment;
                    reads[i] = reverseComplementedString(reads[i]);
                }
            }
            else
            {
                alignments[i] = alignment;
            }
            
            // if (i % 50 == 0) cout << "Aligned read at index " << i << endl;
        }
        
        // The result is written in the fout file
        BufferedWriter fout;
        try
        {
            fout = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file)));
            
            fout.write("@HD\tVN:1.4\tSO:unsorted\n");
            fout.write("@SQ\tSN:" + referenceSeqName + "\tLN:" + reference.length() + "\n");
            
            for (i = 0; i < readsCount; i++)
            {
                for (j = 2; j < 2 + alignments[i][0]; j++)
                    if (alignments[i][2] >= 0)
                    {
                        // print QNAME FLAG RNAME
                        fout.write(readNames[i] + "\t" + 0 + "\t" + referenceSeqName + "\t");
                        // print POS
                        fout.write((alignments[i][j] + 1) + "\t");
                        // print MAPQ
                        fout.write((int)(254 * ((double)alignments[i][1] / maxPossibleDistance)) + "\t");
                        // print CIGAR RNEXT PNEXT TLEN
                        fout.write("0\t*\t0\t0\t");
                        // print SEQ QUAL
                        fout.write(reads[i] + "\t*\t");
                        // print AS:i:<LRD>
                        fout.write("AS:i:" + alignments[i][1] + "\n");
                    }
                    else
                    {
                        // print QNAME FLAG RNAME
                        fout.write(readNames[i] + "\t" + 0 + "\t*\t0\t0\t");
                        // print CIGAR RNEXT PNEXT TLEN
                        fout.write("*\t*\t0\t0\t");
                        // print SEQ QUAL
                        fout.write(reads[i] + "\t*" + "\n");
                    }
            }
            fout.close();
        }
        catch (IOException e)
        {
            System.out.println(e);
            System.exit(1);
        }
    }

    public static void main(String[] args)
    {
        if (args.length != 4)
        {
            System.out.println("Local Rank Distance Sequence Aligner 1.0");
            System.out.println("USAGE: ./lrda <config_file> <reads_file.fastq> <reference_sequence_file.fasta> <output_file.sam>");
            System.out.println("The format of the config file :");
            System.out.println("LINE   1\t<n_gram_size> <max_offset>");
            System.out.println("LINE   2\t<reads_count> <reads_size>");
            System.out.println("LINE   3\t<distance_threshold>");
            System.out.println("LINE   4\t<not_found_ngrams_limit_over_min>");
        }
        else 
        {
            LRDAligner lrd = new LRDAligner(args[0], args[1], args[2]);
            lrd.align(args[3]);
        }
    }
}
