/*
 Rank Distance Sequence Aligner 1.0
 
 This algorithm is based on a hash function that improves the efficiency.
 Aligns a set of reads against a reference DNA sequence using rank distance.
 
 Copyright (C) 2013 Liviu P. Dinu, Radu Tudor Ionescu, Alexandru I. Tomescu
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or any
 later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <fstream>
#include <cmath>
#include <stdlib.h>
#include <time.h>
#include <limits>
#include <vector>

using namespace std;

class RDAligner
{
private:
    char **reads;
    char **readNames;
    
    char *referenceSeq;
    char referenceSeqName[256];

    int readSize, readsCount;
    int referenceSize;
    int maxAllowedDistance;

    vector< vector< vector< vector<int> > > > hash;
    
    int **alignments;
    
    int *offsets;
    int *tempAlignment;
    
    int **A, **C, **G, **T;
    int iA, iC, iG, iT;
    int *kA, *kC, *kG, *kT;
    
    void deleteData();

    int* alignReadAtIndex(int, int, int);
    int buildIndex();
        
public:
    RDAligner(char *, char *, char *);
    ~RDAligner();
    
    int align(char *);
};

// Loads genomes based on a config file given as parameter
RDAligner::RDAligner(char *configFile, char *readsFile, char *referenceFile)
{
	int i, j;
    char ch;
    string dropLine;

    // Read data from config file
	ifstream cfgin;
	cfgin.open(configFile);
	if (!cfgin) 
	{
		cout<<"Unable to open file at path: "<<configFile<<endl;
		throw 0;
	}
    if (!(cfgin>>readsCount))
	{
		cout<<"File at path : "<<configFile<<" is corrupted!"<<endl;
		cfgin.close();
		throw 0;
	}
    if (!(cfgin>>readSize))
	{
		cout<<"File at path : "<<configFile<<" is corrupted!"<<endl;
		cfgin.close();
		throw 0;
	}
    if (!(cfgin>>maxAllowedDistance))
	{
		cout<<"File at path : "<<configFile<<" is corrupted!"<<endl;
		cfgin.close();
		throw 0;
	}
    cfgin.close();
    
    ifstream qin;
    qin.open(readsFile);
    if (!qin)
    {
        cout<<"Unable to open file at path: "<<readsFile<<endl;
        throw 0;
    }
    
    ifstream refin;
    refin.open(referenceFile);
    if (!refin)
    {
        cout<<"Unable to open file at path: "<<referenceFile<<endl;
        qin.close();
        throw 0;
    }
    
    // Get size of reference genome
    do
    {
        getline(refin,dropLine); // drop first line with sequence identifier
    } while (dropLine[0] != '>');
    for (i = 1; dropLine[i] != ' ' && dropLine[i] != '\n' && i < 256; i++)
        referenceSeqName[i-1] = dropLine[i];
    referenceSeqName[i-1] = '\0';
    
    for (referenceSize = 0; refin >> ch; referenceSize++);
    refin.close();
    refin.open(referenceFile);
    if (!refin)
    {
        cout<<"Unable to open file at path: "<<referenceFile<<endl;
        qin.close();
        throw 0;
    }
    
	try
	{
		reads = NULL;
		reads = new char*[readsCount];
        for (i = 0; i < readsCount; i++)
        {
			reads[i] = NULL;
            reads[i] = new char[readSize + 1];
        }
        
        readNames = NULL;
        readNames = new char*[readsCount];
        for (i = 0; i < readsCount; i++)
        {
			readNames[i] = NULL;
            readNames[i] = new char[256];
        }
        
        offsets = NULL;
        offsets = new int[readSize];
        
        tempAlignment = NULL;
        tempAlignment = new int[2 + referenceSize];
        
        iA = 0;
        A = NULL;
        A = new int*[2];
        A[0] = NULL;
        A[0] = new int[referenceSize];
        kA = new int[referenceSize];
        A[1] = NULL;
        A[1] = new int[readSize];
        
        iC = 0;
        C = NULL;
        C = new int*[2];
        C[0] = NULL;
        C[0] = new int[referenceSize];
        kC = new int[referenceSize];
        C[1] = NULL;
        C[1] = new int[readSize];
        
        iG = 0;
        G = NULL;
        G = new int*[2];
        G[0] = NULL;
        G[0] = new int[referenceSize];
        kG = new int[referenceSize];
        G[1] = NULL;
        G[1] = new int[readSize];
        
        iT = 0;
        T = NULL;
        T = new int*[2];
        T[0] = NULL;
        T[0] = new int[referenceSize];
        kT = new int[referenceSize];
        T[1] = NULL;
        T[1] = new int[readSize];
        
		alignments = new int*[readsCount];
		for (i = 0; i < readsCount; i++)
		{
			alignments[i] = NULL;
		}
        
        referenceSeq = NULL;
        referenceSeq = new char[referenceSize + 1];
        
        // Read data from reads (fastq) file
		for (i = 0; i < readsCount; i++)
		{
            do
            {
                getline(qin,dropLine); // drop first line with sequence identifier
            } while (dropLine[0] != '@');
            for (j = 1; dropLine[j] != ' ' && dropLine[j] != '\n' && j < 256; j++)
                readNames[i][j-1] = dropLine[j];
            readNames[i][j-1] = '\0';
            
            for (j = 0; j < readSize;)
                if (qin >> ch)
                {
                    if (ch == 'A' || ch == 'C' || ch =='G' || ch == 'T')
                    {
                        reads[i][j] = ch;
                    }
                    j++;
                }
                else
                {
					cout<<"File at path : "<<readsFile<<" is corrupted!"<<endl;
					throw 0;
				}
            reads[i][readSize] = '\0';
        }
        
        // Read data from reference sequence (fasta) file
        do
        {
            getline(refin,dropLine); // drop first line with sequence identifier
        } while (dropLine[0] != '>');
        
        for (i = 0; i < referenceSize;)
            if (refin >> ch)
            {
                if (ch == 'A' || ch == 'C' || ch =='G' || ch == 'T')
                {
                    referenceSeq[i] = ch;
                }
                i++;
            }
            else
            {
                cout<<"File at path : "<<referenceFile<<" is corrupted!"<<endl;
                throw 0;
            }
        referenceSeq[referenceSize] = '\0';

        qin.close();
        refin.close();
	}
	catch (...)
	{
		deleteData();
        qin.close();
        refin.close();
		throw;
	}
}

RDAligner::~RDAligner()
{
	int i;
	for (i = 0; i < readsCount; i++)
		delete[] reads[i];
    delete[] reads;
    reads = NULL;
    
    for (i = 0; i < readsCount; i++)
		delete[] readNames[i];
    delete[] readNames;
    readNames = NULL;
    
    delete[] offsets;
    offsets = NULL;
    
    delete[] tempAlignment;
    tempAlignment = NULL;
    
	for (i = 0; i < readsCount; i++)
		delete[] alignments[i];
	delete[] alignments;
	alignments = NULL;
    
	delete[] referenceSeq;
	referenceSeq = NULL;
    
    delete[] A[0];
    delete[] A[1];
    delete[] A;
    delete[] kA;
    
    delete[] C[0];
    delete[] C[1];
    delete[] C;
    delete[] kC;
    
    delete[] G[0];
    delete[] G[1];
    delete[] G;
    delete[] kG;
    
    delete[] T[0];
    delete[] T[1];
    delete[] T;
    delete[] kT;
    
	readsCount = readSize = referenceSize = maxAllowedDistance = 0;
}

void RDAligner::deleteData()
{
    int i;
	if (reads != NULL)
	{
		for (i = 0; i < readsCount; i++)
            delete[] reads[i];
        delete[] reads;
        reads = NULL;
	}
    if (readNames != NULL)
    {
        for (i = 0; i < readsCount; i++)
            delete[] readNames[i];
        delete[] readNames;
        readNames = NULL;
    }
    if (offsets != NULL)
    {
        delete[] offsets;
        offsets = NULL;
    }
    if (tempAlignment != NULL)
    {
        delete[] tempAlignment;
        tempAlignment = NULL;
    }
	if (alignments != NULL)
	{
		for (i = 0; i < readsCount; i++)
            delete[] alignments[i];
        delete[] alignments;
        alignments = NULL;
	}
	if (referenceSeq != NULL)
	{
		delete[] referenceSeq;
		referenceSeq = NULL;
	}
    if (A != NULL)
    {
        if (A[0] != NULL) delete[] A[0];
        if (A[1] != NULL) delete[] A[1];
        delete[] A;
        A = NULL;
    }
    if (C != NULL)
    {
        if (C[0] != NULL) delete[] C[0];
        if (C[1] != NULL) delete[] C[1];
        delete[] C;
        C = NULL;
    }
    if (G != NULL)
    {
        if (G[0] != NULL) delete[] G[0];
        if (G[1] != NULL) delete[] G[1];
        delete[] G;
        G = NULL;
    }
    if (T != NULL)
    {
        if (T[0] != NULL) delete[] T[0];
        if (T[1] != NULL) delete[] T[1];
        delete[] T;
        T = NULL;
    }
    if (kA != NULL)
    {
        delete[] kA;
        kA = NULL;
    }
    if (kC != NULL)
    {
        delete[] kC;
        kC = NULL;
    }
    if (kG != NULL)
    {
        delete[] kG;
        kG = NULL;
    }
    if (kT != NULL)
    {
        delete[] kT;
        kT = NULL;
    }
}

// Aligns a read against the reference DNA sequence
int* RDAligner::alignReadAtIndex(int readIndex, int useReverseComplement, int bestDist)
{
    int i, k, pos;
    int dist = 0;
    int jA = 0, jC = 0, jG = 0, jT = 0;
    int minA, minC, minG, minT;
    int numberOfAs = 0, numberOfCs = 0, numberOfGs = 0, numberOfTs = 0;

    for (i = 0; i < readSize; i++)
    {
        A[1][i] = 0;
        C[1][i] = 0;
        G[1][i] = 0;
        T[1][i] = 0;
    }
    
    if (useReverseComplement)
    {
        for (i = readSize - 1; i >= 0; i--)
        {
            if (reads[readIndex][i] == 'A')
            {
                T[1][jT] = i;
                jT++;
                numberOfTs++;
            }
            else if (reads[readIndex][i] == 'C')
            {
                G[1][jG] = i;
                jG++;
                numberOfGs++;
            }
            else if (reads[readIndex][i] == 'G')
            {
                C[1][jC] = i;
                jC++;
                numberOfCs++;
            }
            else if (reads[readIndex][i] == 'T')
            {
                A[1][jA] = i;
                jA++;
                numberOfAs++;
            }
        }
    }
    else
    {
        for (i = 0; i < readSize; i++)
        {
            if (reads[readIndex][i] == 'A')
            {
                A[1][jA] = i;
                jA++;
                numberOfAs++;
            }
            else if (reads[readIndex][i] == 'C')
            {
                C[1][jC] = i;
                jC++;
                numberOfCs++;
            }
            else if (reads[readIndex][i] == 'G')
            {
                G[1][jG] = i;
                jG++;
                numberOfGs++;
            }
            else if (reads[readIndex][i] == 'T')
            {
                T[1][jT] = i;
                jT++;
                numberOfTs++;
            }
        }
    }
    
    tempAlignment[0] = 1;
    tempAlignment[1] = bestDist;
    tempAlignment[2] = -1;
    
    for (int slotAShift = 0; (slotAShift < readSize) and (slotAShift * readSize <= bestDist); slotAShift++)
    {
        for (int signA = -1; signA < 2; signA = signA + 2)
        {
            if ((signA == 1) and (slotAShift == 0)) break;
            for (int slotCShift = 0; (slotCShift < readSize) and (slotCShift * readSize + slotAShift * readSize <= bestDist); slotCShift++)
            {
                for (int signC = -1; signC < 2; signC = signC + 2)
                {
                    if ((signC == 1) and (slotCShift == 0)) break;
                    for (int slotGShift = 0; (slotGShift < readSize) and (slotGShift * readSize + slotCShift * readSize + slotAShift * readSize <= bestDist); slotGShift++)
                    {
                        for (int signG = -1; signG < 2; signG = signG + 2)
                        {
                            if ((signG == 1) and (slotGShift == 0)) break;
                            
                            int slotA = numberOfAs + signA * slotAShift;
                            int slotC = numberOfCs + signC * slotCShift;
                            int slotG = numberOfGs + signG * slotGShift;
                            if ((slotA >= 0) and (slotA < readSize) and (slotC >= 0) and (slotC < readSize) and (slotG >= 0) and (slotG < readSize))
                            {
                                for (vector<int>::iterator vit = hash[slotA][slotC][slotG].begin(); vit != hash[slotA][slotC][slotG].end(); ++vit)
                                {
                                    //pos = HAvalues[k];
                                    pos = *vit;
                                    // compute RD between read and reference starting at index *sit
                                    dist = 0;
                                    
                                    /*
                                    minA = (jA < iA - kA[pos]) ? jA : iA - kA[pos];
                                    minC = (jC < iC - kC[pos]) ? jC : iC - kC[pos];
                                    minG = (jG < iG - kG[pos]) ? jG : iG - kG[pos];
                                    minT = (jT < iT - kT[pos]) ? jT : iT - kT[pos];
                                    */
                                    minA = (jA < kA[pos + readSize] - kA[pos]) ? jA : kA[pos + readSize] - kA[pos];
                                    minC = (jC < kC[pos + readSize] - kC[pos]) ? jC : kC[pos + readSize] - kC[pos];
                                    minG = (jG < kG[pos + readSize] - kG[pos]) ? jG : kG[pos + readSize] - kG[pos];
                                    minT = (jT < kT[pos + readSize] - kT[pos]) ? jT : kT[pos + readSize] - kT[pos];

                                    //if (abs(kA[pos + readSize] - kA[pos] - jA) != slotAShift)
                                    //    cout << "error" << endl;
                                    
                                    for (i = 0; i < minA; i++)
                                        dist = dist + abs((A[0][i + kA[pos]] - pos) - A[1][i]);
                                    if (dist > bestDist) continue;
                                    for (i = 0; i < minC; i++)
                                        dist = dist + abs((C[0][i + kC[pos]] - pos) - C[1][i]);
                                    if (dist > bestDist) continue;
                                    for (i = 0; i < minG; i++)
                                        dist = dist + abs((G[0][i + kG[pos]] - pos) - G[1][i]);
                                    if (dist > bestDist) continue;
                                    for (i = 0; i < minT; i++)
                                        dist = dist + abs((T[0][i + kT[pos]] - pos) - T[1][i]);
                                    if (dist > bestDist) continue;
                                    
                                    dist = dist + slotAShift * readSize;
                                    dist = dist + slotCShift * readSize;
                                    dist = dist + slotGShift * readSize;
                                    dist = dist + abs(kT[pos + readSize] - kT[pos] - jT) * readSize;
                                    
                                    // cout << "Dist at " << k << " : " << dist << endl;
                                    
                                    if (dist < bestDist)
                                    {
                                        tempAlignment[0] = 1;
                                        tempAlignment[1] = dist;
                                        tempAlignment[2] = pos;
                                        bestDist = dist;
                                    }
                                    else if (dist == bestDist)
                                    {
                                        if (tempAlignment[2] >= 0) tempAlignment[0]++;
                                        tempAlignment[1 + tempAlignment[0]] = pos;
                                    }
                                }
                            }
                        }
                    }                
                }
            }
        }
    }
    
    int *alignment = NULL;
    try
    {
        alignment = new int[2 + tempAlignment[0]];
        for (i = 0; i < 2 + tempAlignment[0]; i++)
            alignment[i] = tempAlignment[i];
    }
    catch (...)
    {
        return NULL;
    }
    return alignment;
}

// Build the A-index over the text
int RDAligner::buildIndex()
{
    int i;
    int numberOfAs = 0, numberOfCs = 0, numberOfGs = 0, numberOfTs = 0;
    hash = vector<vector<vector<vector<int> > > >(readSize, vector<vector<vector<int> > >(readSize, vector<vector<int> >(readSize,vector<int>(0))));

    for (i = 0; i < readSize; i++)
    {
        if (referenceSeq[i] == 'A')
        {
            numberOfAs++;
        }
        else if (referenceSeq[i] == 'C')
        {
            numberOfCs++;
        }
        else if (referenceSeq[i] == 'G')
        {
            numberOfGs++;
        }
        else if (referenceSeq[i] == 'T')
        {
            numberOfTs++;
        }
        
    }
    hash[numberOfAs][numberOfCs][numberOfGs].push_back(0);
    
	for (i = readSize; i < referenceSize; i++)
	{
        if (referenceSeq[i] == 'A')
        {
            numberOfAs++;
        }
        else if (referenceSeq[i] == 'C')
        {
            numberOfCs++;
        }
        else if (referenceSeq[i] == 'G')
        {
            numberOfGs++;
        }
        else if (referenceSeq[i] == 'T')
        {
            numberOfTs++;
        }

        if (referenceSeq[i - readSize] == 'A')
        {
            numberOfAs--;
        }
        else if (referenceSeq[i - readSize] == 'C')
        {
            numberOfCs--;
        }
        else if (referenceSeq[i - readSize] == 'G')
        {
            numberOfGs--;
        }
        else if (referenceSeq[i - readSize] == 'T')
        {
            numberOfTs--;
        }
        
        hash[numberOfAs][numberOfCs][numberOfGs].push_back(i - readSize + 1);
    }
    
    for (i = 0; i < referenceSize; i++)
    {
        A[0][i] = 0;
        C[0][i] = 0;
        G[0][i] = 0;
        T[0][i] = 0;
    }
    
    for (i = 0; i < referenceSize; i++)
	{
        if (referenceSeq[i] == 'A')
        {
            A[0][iA] = i;
            iA++;
        }
        else if (referenceSeq[i] == 'C')
        {
            C[0][iC] = i;
            iC++;
        }
        else if (referenceSeq[i] == 'G')
        {
            G[0][iG] = i;
            iG++;
        }
        else if (referenceSeq[i] == 'T')
        {
            T[0][iT] = i;
            iT++;
        }
    }

    kA[0] = kC[0] = kG[0] = kT[0] = 0;
    for (int k = 1; k <= referenceSize; k++)
    {
        kA[k] = kA[k-1];
        kC[k] = kC[k-1];
        kG[k] = kG[k-1];
        kT[k] = kT[k-1];
        if (referenceSeq[k - 1] == 'A') kA[k]++;
        else if (referenceSeq[k - 1] == 'C') kC[k]++;
        else if (referenceSeq[k - 1] == 'G') kG[k]++;
        else if (referenceSeq[k - 1] == 'T') kT[k]++;
    }
    
    return 1;
}

// Aligns all reads against a reference DNA sequence using RD and writes the results in the file given as parameter
int RDAligner::align(char *file)
{
	if (readsCount < 0) return 0;
	int i, j;
    
    int maxPossibleDistance = readSize * readSize;
    int maxDistance = (maxAllowedDistance < maxPossibleDistance) ? maxAllowedDistance : maxPossibleDistance;

    int maxAcceptableDistance = (maxAllowedDistance < 3 * readSize) ? maxAllowedDistance : 3 * readSize;

    cout << "Building the indices..." << endl;
    buildIndex();
	for (i = 0; i < readsCount; i++)
	{
        int *alignment;
        int *reverseComplementAlignment;
        
        alignment = alignReadAtIndex(i, 0, maxDistance);
        reverseComplementAlignment = alignReadAtIndex(i, 1, alignment[1]);
        
        if (alignment[1] <= reverseComplementAlignment[1])
        {
            alignments[i] = alignment;
        }
        else
        {
            alignments[i] = reverseComplementAlignment;
        }
        
        // if (i % 100 == 0) cout << "Aligned read at index " << i << endl;
	}
    
    // The result is written in the fout file
    ofstream fout;
	fout.open(file);
	if (!fout)
	{
		cout << "Unable to open file at path: " << file << endl;
		return 0;
	}
    
    fout << "@HD\tVN:1.4\tSO:unsorted" << endl;
    fout << "@SQ\tSN:" << referenceSeqName <<"\tLN:" << referenceSize << endl;
	for (i = 0; i < readsCount; i++)
	{
		for (j = 2; j < 2 + alignments[i][0]; j++)
            if (alignments[i][2] >= 0)
            {
                // print QNAME FLAG RNAME
                fout << readNames[i] << "\t" << 0 << "\t" << referenceSeqName << "\t";
                // print POS
                fout << (alignments[i][j] + 1) << "\t";
                // print MAPQ
                fout << int(254 * ((double)alignments[i][1] / maxPossibleDistance)) << "\t";
                // print CIGAR RNEXT PNEXT TLEN
                fout << "0\t*\t0\t0\t";
                // print SEQ QUAL
                fout << reads[i] << "\t*\t";
                // print AS:i:<RD>
                fout << "AS:i:" << alignments[i][1] << endl;
            }
            else
            {
                // print QNAME FLAG RNAME
                fout << readNames[i] << "\t" << 0 << "\t*\t0\t0\t";
                // print CIGAR RNEXT PNEXT TLEN
                fout << "*\t*\t0\t0\t";
                // print SEQ QUAL
                fout << reads[i] << "\t*" << endl;
            }
	}
	fout.close();
	return 1;
}

int main(int argc, char *argv[])
{
    if (argc != 5)
    {
        cout<<"Rank Distance Sequence Aligner 1.0"<<endl;
        cout<<"USAGE: ./rda <config_file> <reads_file.fastq> <reference_sequence_file.fasta> <output_file.sam>"<<endl;
        cout<<"The format of the config file :"<<endl;
        cout<<"LINE   1\t<reads_count> <reads_size>"<<endl;
        cout<<"LINE   2\t<distance_threshold>"<<endl;
        return 1;
    }
    else 
    {
        RDAligner *rd;
        try 
        {
            rd = new RDAligner(argv[1], argv[2], argv[3]);
            
            if (!rd->align(argv[4]))
            {
                return 1;
            }
            
            delete rd;
        }
        catch(...) 
        {
            return 1;
        }  	
    }
    return 0;
}
