/*
 SAM Evaluation Tool 1.0
 
 Evaluates a *.sam file and generates an overall score. 
 A lower score indicates a better similarity between the reads and the reference.
 
 Copyright (C) 2013  Liviu P. Dinu, Radu Tudor Ionescu, Alexandru I. Tomescu
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or any
 later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <fstream>
#include <cmath>
#include <stdlib.h>
#include <time.h>
#include <limits>

using namespace std;

class SAMeval
{
private:
    char distType;
    char fieldPrefix[6];
    long maxDist;
    long readsCount;
    
    void deleteData();
        
public:
    SAMeval(char *);
    ~SAMeval();
    
    int eval(char *);
};

// Loads genomes based on a config file given as parameter
SAMeval::SAMeval(char *configFile)
{
	// int i, j;
    // char ch;
    string dropLine;

    // Read data from config file
	ifstream cfgin;
	cfgin.open(configFile);
	if (!cfgin) 
	{
		cout<<"Unable to open file at path: "<<configFile<<endl;
		throw 0;
	}
    if (!(cfgin>>distType))
	{
		cout<<"File at path : "<<configFile<<" is corrupted!"<<endl;
		cfgin.close();
		throw 0;
	}
    if (!(cfgin>>maxDist))
	{
		cout<<"File at path : "<<configFile<<" is corrupted!"<<endl;
		cfgin.close();
		throw 0;
	}
    if (!(cfgin>>readsCount))
	{
		cout<<"File at path : "<<configFile<<" is corrupted!"<<endl;
		cfgin.close();
		throw 0;
	}
    if (distType == 'e') strcpy(fieldPrefix,"NM:i:");
    else if (distType == 'o') strcpy(fieldPrefix,"AS:i:");
    else if (distType == 's') strcpy(fieldPrefix,", ");
    fieldPrefix[5] = '\0';
    
    cfgin.close();
}

SAMeval::~SAMeval()
{
    maxDist = readsCount = 0;
}

void SAMeval::deleteData()
{
    // nothing to delete
}

// Evaluates all alignments from a .sam file and computes a score 
int SAMeval::eval(char *file)
{
    ifstream fin;
	fin.open(file);
	if (!fin)
	{
		cout << "Unable to open file at path: " << file << endl;
		return 0;
	}
    
    int i;
    int alignedReadsCount = 0;
    int dist;
    double score;
    double distSum = 0.0;
    
    string row;
    char cRow[1024];
    char *foundPtr;
    
    for (i = 0; !fin.eof(); i++)
    {
        getline(fin,row); 
        if (row[0] != '@') // drop first lines with meta-information
        {
            strcpy (cRow, row.c_str());
            
            foundPtr = NULL;
            foundPtr = strstr(cRow, fieldPrefix);
            if (foundPtr != NULL)
            {
                if (distType == 'e') sscanf(foundPtr, "NM:i:%d", &dist);
                else if (distType == 'o') sscanf(foundPtr, "AS:i:%d", &dist);
                else if (distType == 's')
                {
                    sscanf(foundPtr, ", %d", &dist);
                    dist = 1000 - dist;
                }
                // cout << "Dist at index " << i << " : " << dist << endl;
                if (dist <= maxDist)
                {
                    distSum += dist;
                    alignedReadsCount++;
                }
            }
        }
    }
    fin.close();

    if (alignedReadsCount > readsCount) readsCount = alignedReadsCount;
    score = (distSum / alignedReadsCount) * ((float)(1.0 + readsCount - alignedReadsCount)) / (float)readsCount;

    char scoreFile[256];
    sprintf(scoreFile, "%s-score.txt", file);
    ofstream fout;
	fout.open(scoreFile);
	if (!fout)
	{
		cout << "Unable to open file at path: " << file << endl;
		return 0;
	}
    
    fout << "Score : " << score << endl;
    fout << "Mean distance : " << (distSum / alignedReadsCount) << endl;
    fout << "Aligned reads : " << alignedReadsCount << endl;
	fout.close();
    
    cout << "Results for file : " << file << endl;
    cout << "Score : " << score << endl;
    cout << "Mean distance : " << (distSum / alignedReadsCount) << endl;
    cout << "Aligned reads : " << alignedReadsCount << endl;
	
	return 1;
}

int main(int argc, char *argv[])
{
    if (argc != 3)
    {
        cout<<"USAGE: ./sameval <config_file.txt> <input_file.sam>"<<endl;
        cout<<"The format of the config file :"<<endl;
        cout<<"LINE   1\t 'e' for edit distance (looks for NM field)"<<endl;
        cout<<"        \t 'o' for other distance (looks for AS field)"<<endl;
        cout<<"        \t 's' for score (looks for AS field)"<<endl;
        cout<<"LINE   2\t <max_dist> <reads_count>"<<endl;
        return 1;
    }
    else 
    {
        SAMeval *sameval;
        try 
        {
            sameval = new SAMeval(argv[1]);
            
            if (!sameval->eval(argv[2]))
            {
                cout<<"Press any key and ENTER to exit..."<<endl;
                return 1;
            }
            
            delete sameval;
        }
        catch(...) 
        {
            cout<<"Press any key and ENTER to exit..."<<endl;
            return 1;
        }  	
    }
    return 0;
}
